"""
A module containing exceptions for swk.

swk - A tiny extendable utility for running commands against multiple hosts.

Copyright (C) 2016  Pavel "trueneu" Gurkov

see swk/main.py for more information on License and contacts
"""


class OverriddenCommandError(Exception):
    def __init__(self, message):
        super(OverriddenCommandError, self).__init__(message)


class OverriddenParserError(Exception):
    def __init__(self, message):
        super(OverriddenParserError, self).__init__(message)


class ExpandingHostlistError(Exception):
    def __init__(self, message):
        super(ExpandingHostlistError, self).__init__(message)


class ConfigNotFoundError(Exception):
    def __init__(self, message):
        super(ConfigNotFoundError, self).__init__(message)



"""
A module containing interaction layer with pypsi shell.

swk - A tiny extendable utility for running commands against multiple hosts.

Copyright (C) 2016  Pavel "trueneu" Gurkov

see swk/main.py for more information on License and contacts
"""

from pypsi.shell import Shell
from pypsi.core import Command
import sys
import logging
from swk import classes
from swk import exceptions
from pypsi.commands.exit import ExitCommand
from pypsi.commands.pwd import PwdCommand
from pypsi.commands.chdir import ChdirCommand
from pypsi.commands.system import SystemCommand
import os
from pypsi.plugins.history import HistoryCommand, HistoryPlugin
from pypsi.core import Plugin
from pypsi.cmdline import StringToken, WhitespaceToken, OperatorToken
from pypsi.os import find_bins_in_path


class SWKShellCmdlinePP(Plugin):
    def __init__(self, preprocess=80, postprocess=None, **kwargs):
        super(SWKShellCmdlinePP, self).__init__(preprocess=preprocess,
                                                postprocess=postprocess, **kwargs)

    def setup(self, shell):
        pass

    def on_tokenize(self, shell, tokens, origin):
        if origin != 'input':
            return tokens

        # I really have to return here to Meily's variant with detecting "the end of command" case
        # collect cmd, hostlist, and pass the quoted argument
        # and then shlex it on the other side with unquoting there

        result = []
        i = 0

        cmd = None
        hostlist = None
        arg = ''
        i = 0
        for (i, token) in enumerate(tokens):
            if isinstance(token, StringToken):
                # We only really care about string tokens
                if cmd:
                    if cmd.text not in shell._swk_instance._available_commands.keys():
                        result.append(token)
                    else:
                        if hostlist:
                            arg += "{quote}{s}{quote}".format(
                                quote=token.quote or '',
                                s=token.text.replace('\\', '\\\\\\')
                            )
                        else:
                            result.append(token)
                            hostlist = token
                else:
                    # The first string token we see is the actual command we
                    # need to execute
                    result.append(token)
                    cmd = token

            elif isinstance(token, WhitespaceToken):
                if arg:
                    # append the whitespace
                    arg += ' '
                elif cmd:
                    # We need whitespace between the command and the arguments
                    # so add it here
                    result.append(token)

            elif isinstance(token, OperatorToken):
                if token.operator in ('&&', '|', '||', ';', '>', '<'):
                    # We found a chainning operator so start over
                    cmd = None
                    hostlist = None
                    if arg:
                        result.append(StringToken(i, arg))
                        arg = ''
                result.append(token)

            else:
                # Unknown token
                result.append(token)

        if arg:
            result.append(StringToken(i+1, arg))

        return result


help_command_name = 'help'
swk_shell_prompt = "swk> "


class SWKChdirCommand(ChdirCommand):
    def run(self, shell, args):
        super(SWKChdirCommand, self).run(shell, args)
        shell.prompt = swk_shell_prompt.format(os.getcwd())


class SWKCommand(Command):
    def __init__(self, command_name, command_help, command_module, swk_instance):
        super(SWKCommand, self).__init__(command_name, usage=command_help)
        self._command_name = command_name
        self._command_help = command_help
        self._command_module = self._command_executer_class = command_module
        self._command_executer_name = command_module.__name__
        self._hostlist = ""
        self._command_args = list()
        self._swk_instance = swk_instance
        self._expanded_hostlist = list()
        self._cwd = ""

    def _die(self, diemsg):
        logging.error(diemsg)
        sys.stderr.write(diemsg + '\n')

    def run(self, shell, args):
        self._cwd = os.getcwd()

        if self._command_module.requires_hostlist(self._command_name):
            try:
                self._hostlist = args[0]
            except IndexError:
                self._die("Hostlist is required for command {0}, but is not provided".format(self._command_name))
                return
            self._command_args = args[1:]

            self._swk_instance._hostlist = self._hostlist
            try:
                self._expanded_hostlist = self._swk_instance._expand_hostlist()
            except classes.SWKParsingError as e:
                self._die("Parser error: {0}".format(str(e)))
                return
            except exceptions.ExpandingHostlistError as e:
                self._die("Expanding hostlist expression error: {0}".format(str(e)))

        else:
            self._command_args = args

        self._swk_instance._config[self._command_executer_name] = \
            self._swk_instance._update_config(self._command_executer_name,
                                              hostlist=self._expanded_hostlist,
                                              command=self._command_name,
                                              command_args=self._command_args,
                                              swk_dir=self._swk_instance._swk_dir,
                                              swk_path=self._swk_instance._swk_path,
                                              cwd=self._cwd,
                                              cache_directory=self._swk_instance._cache_directory_expanded,
                                              called_from_shell=True)

        logging.info("Executing command with config: {0}".format(self._swk_instance._config[self._command_executer_name]))
        obj = self._command_executer_class(**self._swk_instance._config[self._command_executer_name])
        try:
            obj.run_command()
        except classes.SWKCommandError as e:
            self._die("Command error: {0}".format(str(e)))


class SWKShell(Shell):
    exit_cmd = ExitCommand()
    pwd_cmd = PwdCommand()
    cd_cmd = SWKChdirCommand()
    system_cmd = SystemCommand(name='sys')
    history_plugin = HistoryPlugin(history_cmd='hist')
    history_cmd = HistoryCommand(name='hist')
    enabled_builtin_pypsi_cmds = {'exit': exit_cmd, 'pwd': pwd_cmd, 'sys': system_cmd, 'cd': cd_cmd, 'hist': history_cmd}
    history_file_path = "~/.swk/.history"
    history_file_path_expanded = os.path.expanduser(history_file_path)
    swk_shell_preprocessor_plugin = SWKShellCmdlinePP()

    def __init__(self, swk_instance):
        try:
            _, columns = os.popen('stty size 2>/dev/null', 'r').read().split()
        except ValueError:  # we're not running in a terminal
            columns = 80
        super(SWKShell, self).__init__(shell_name='swk', width=int(columns))
        self.fallback_cmd = self.system_cmd
        self.prompt = swk_shell_prompt.format(os.getcwd())
        self._sys_bins = None
        self._swk_instance = swk_instance

    def on_cmdloop_begin(self):
        if os.path.exists(self.history_file_path_expanded):
            self.history_cmd.run(self, ['load', '{0}'.format(self.history_file_path_expanded)])

    def on_cmdloop_end(self):
        self.history_cmd.run(self, ['save', '{0}'.format(self.history_file_path_expanded)])

    def get_command_name_completions(self, prefix):
        if not self._sys_bins:
            self._sys_bins = find_bins_in_path()

        return sorted(
            [name for name in self.commands if name.startswith(prefix)] +
            [name for name in self._sys_bins if name.startswith(prefix)]
        )


class SWKShellPrepare:
    def __init__(self, swk_instance):
        for command_name, command_module in swk_instance._available_commands.items():
            setattr(SWKShell, command_name, SWKCommand(command_name, 'no help yet', command_module, swk_instance))
        setattr(SWKShell, help_command_name, SWKShellHelp(swk_instance))


class SWKShellHelp(Command):
    def __init__(self, swk_instance):
        super(SWKShellHelp, self).__init__(help_command_name)
        self._commands_help_message = swk_instance._commands_help_message
        self._parsers_help_message = swk_instance._parsers_help_message
        self._swk_instance = swk_instance

    def run(self, shell, args):
        builtin_commands_string = ""
        for builtin_command in sorted(shell.enabled_builtin_pypsi_cmds.keys()):
            builtin_commands_string += builtin_command + ', '
        builtin_commands_string = builtin_commands_string[:-2]
        if len(args) == 0:
            sys.stdout.write("{0}{1}".format(self._commands_help_message, self._parsers_help_message))
            sys.stdout.write("\nBuilt-in commands available: {0}\n".format(builtin_commands_string))
            sys.stdout.write("\nFor verbose help, please run 'help <command_name>' or 'help <parser_modifier>'\n")
            return
        help_topic_name = args[0]
        if help_topic_name in shell.enabled_builtin_pypsi_cmds.keys():
            sys.stdout.write("{0}\n".format(shell.enabled_builtin_pypsi_cmds[help_topic_name].brief))
            return
        try:
            sys.stdout.write(self._swk_instance._available_commands[help_topic_name].get_command_help(help_topic_name))
        except KeyError:
            try:
                sys.stdout.write(self._swk_instance._available_parsers[help_topic_name].get_parser_help(help_topic_name))
            except KeyError:
                sys.stderr.write("Sorry, no help on this topic\n")





"""
Main program module.

swk - A tiny extendable utility for running commands against multiple hosts.

Copyright (C) 2016  Pavel "trueneu" Gurkov

see swk/main.py for more information on License and contacts
"""

import logging
import os
import glob

from swk import classes
import inspect
from swk import exceptions
import argparse
import configparser
import sys
import exrex
import pkg_resources
from swk import version
import shutil
from swk import check_updates
import datetime

shell_mode_off = False
try:
    import swk.shell
except SyntaxError:
    #logging.warning("swk shell mode has been disabled. Seems like you're using Python 2")
    shell_mode_off = True
except ImportError:
    #logging.warning("swk shell mode has been disabled. Seems like you haven't installed pypsi package")
    shell_mode_off = True


class SwissKnife(object):
    _version = version.__version__

    swk_plugin_dir_default = "plugins"
    _swk_config_path = "~/.swk/"
    _swk_config_filename = "swk.ini"
    _swk_config_full_path = os.path.join(os.path.expanduser(_swk_config_path), _swk_config_filename)
    _swk_package_name = 'swk'
    _swk_check_updates_marker_filename = 'checked_updates'
    _swk_check_updates_marker_full_path = os.path.join(os.path.expanduser(_swk_config_path),
                                                       _swk_check_updates_marker_filename)
    _swk_check_updates_period = 60 * 60 * 24

    def _write_default_config(self):
        if not os.path.isdir(os.path.dirname(self._swk_config_full_path)):
            try:
                os.mkdir(os.path.dirname(self._swk_config_full_path))
            except IOError:
                msg = "Couldn't create directory {path}, aborting.".format(
                    path=os.path.dirname(self._swk_config_full_path))
                self._die(msg)

        if not os.path.isfile(self._swk_config_full_path):
            try:
                shutil.copy(os.path.join(os.path.dirname(classes.__file__), self._swk_config_filename),
                            self._swk_config_full_path)
            except IOError:
                msg = "Couldn't create default config at {path}, aborting.".format(path=self._swk_config_full_path)
                self._die(msg)

    def _check_updates(self, package_version_dict):
        now = datetime.datetime.utcnow()
        now_timestamp = (now - datetime.datetime(1970, 1, 1)).total_seconds()
        try:
            check_updates_marker_mtime = os.path.getmtime(self._swk_check_updates_marker_full_path)
        except (IOError, OSError):
            check_updates_marker_mtime = now_timestamp - (self._swk_check_updates_period + 1)
        if now_timestamp - check_updates_marker_mtime > self._swk_check_updates_period:
            try:
                cheese_shop = check_updates.CheeseShop()
            except:
                msg = "Couldn't run check for new versions. Please check your Internet settings or turn " \
                      "checking for updates off" \
                      "in {0}.".format(self._swk_config_filename)
                logging.error(msg)
                sys.stderr.write(msg + '\n')
                sys.stderr.flush()
                return
            for k, v in package_version_dict.items():
                package_name = k
                installed_version = v
                try:
                    last_version = cheese_shop.package_releases(package_name)[0]
                except:
                    msg = "Couldn't run check for new versions of {package}. " \
                          "Please check your Internet settings or turn " \
                          "checking for updates off" \
                          "in {conf_file}.".format(conf_file=self._swk_config_filename, package=package_name)
                    logging.error(msg)
                    sys.stderr.write(msg + '\n')
                    sys.stderr.flush()
                    last_version = installed_version
                if check_updates.get_highest_version([installed_version, last_version]) != installed_version:
                    sys.stderr.write("You're using {package} v{old}, but v{new} is available! Please upgrade.\n".format(
                        old=installed_version, new=last_version, package=package_name
                    ))
                    logging.info("You're using {package} v{old}, but v{new} is available! Please upgrade.\n".format(
                        old=installed_version, new=last_version, package=package_name
                    ))

            with open(self._swk_check_updates_marker_full_path, mode='w'):
                pass

    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, "_{0}".format(k), v)

        self._write_default_config()
        self._config = self._read_config()
        
        self._logging_init()

        package_version_dict = {self._swk_package_name: self._version}

        self._disabled_plugins = [plugin for plugin in self._config["Main"].get("disabled_plugins", "").split()]

        self._cache_directory_expanded = os.path.abspath(os.path.expanduser(self._config["Main"].get("cache_directory",
                                                                                                  "~/.swk")))
        self._cache_directory_init()

        self._swk_plugins_dirs = [os.path.expanduser(x) for x in self._config["Main"].get("plugins_directories", "").split()]
        self._swk_plugins_dirs.append("{0}/{1}".format(self._swk_dir, self.swk_plugin_dir_default))
        self._plugin_modules, self._plugin_command_modules, self._plugin_parser_modules,\
            modules_package_version_dict = self._modules_import()

        package_version_dict.update(modules_package_version_dict)
        if self._config["Main"].get("check_for_updates", "no") == "yes":
            self._check_updates(package_version_dict)

        self._available_commands, self._available_parsers = self._get_available_commands_and_parsers()
        self._commands_help_message, self._parsers_help_message, \
            self._commands_help_string, self._parsers_help_string = self._form_commands_and_parsers_help()

        self._config = self._add_empty_sections_to_config()  # now that we know what's imported

        self._args = self._parse_args()

        self._command = self._args["command"]

        if self._command == 'shell':
            pass
        else:
            """almost all this stuff below we do based on arguments we got so it gotta be excluded from shell variant"""
            self._command_executer_class = self._find_command_executer_class()
            self._command_executer_name = self._command_executer_class.__name__
            self._command_requires_hostlist = self._command_executer_class.requires_hostlist(self._args["command"])
            self._args = self._arguments_magic()

            self._command_args = self._args["command_args"]
            self._hostlist = self._args["hostlist"]



    def _logging_init(self):
        loglevel_string = self._config["Main"].get("loglevel", "error")
        try:
            loglevel = {"debug": logging.DEBUG,
                        "info": logging.INFO,
                        "warning": logging.WARNING,
                        "error": logging.ERROR,
                        "critical": logging.CRITICAL}[loglevel_string]
        except KeyError:
            loglevel = logging.ERROR
        if loglevel == logging.DEBUG:
            self._dbg_prints = True
        else:
            self._dbg_prints = False
        logfile = os.path.expanduser(self._config["Main"].get("logfile", "~/.swk/swk.log"))
        logging.basicConfig(filename=logfile, filemode='a', level=loglevel,
                            format='[%(asctime)s] %(levelname)s : %(message)s', datefmt='%Y-%m-%d %H:%M:%S')
        logging.debug("swk started")

    def _cache_directory_init(self):
        if not os.path.exists(self._cache_directory_expanded):
            try:
                os.mkdir(self._cache_directory_expanded)
            except OSError:
                self._die("Couldn't create {0} directory. Check permissions".format(self._cache_directory_expanded))


    @staticmethod
    def _die(diemsg):
        logging.error(diemsg)
        sys.stderr.write(diemsg + '\n')
        logging.debug("swk died")
        exit(2)

    def _modules_import(self):

        module_filenames = list()
        plugin_modules = list()
        plugin_command_modules = list()
        plugin_parser_modules = list()
        package_version_dict = dict()
        oldcwd = os.getcwd()

        # directory imports

        for swk_plugins_dir in self._swk_plugins_dirs:
            try:
                swk_plugins_dir_abspath = os.path.abspath(swk_plugins_dir)
                os.chdir(swk_plugins_dir_abspath)
                module_filenames = glob.glob("*.py")
            except OSError:
                self._die("{0} does not exist.".format(swk_plugins_dir_abspath))

            """import all the plugin modules and put them into list"""

            for module_filename in module_filenames:
                module_name, _, _ = module_filename.rpartition('.py')
                if module_name in self._disabled_plugins or module_name in ['__init__']:
                    continue

                try:
                    sys.path.append(os.getcwd())
                    module = __import__(module_name)
                except ImportError as e:
                    # self._die("Couldn't import module {0}: {1}.".format(module_name, str(e)))
                    logging.error("Couldn't import module {0}: {1}.".format(module_name, str(e)))
                plugin_modules.extend([(name, obj) for (name, obj) in inspect.getmembers(module)
                                       if inspect.isclass(obj) and issubclass(obj, classes.SWKPlugin)])
        os.chdir(oldcwd)

        # entry_points imports
        for entry_point in pkg_resources.iter_entry_points(group='swk_plugin', name=None):
            #first variant
            module_name = entry_point.module_name

            if module_name in self._disabled_plugins or module_name in ['__init__']:
                continue
            try:
                module = __import__(module_name, fromlist=[module_name[:module_name.rfind('.')]])
            except ImportError as e:
                # self._die("Couldn't import module {0}: {1}.".format(module_name, str(e)))
                logging.error("Couldn't import module {0}: {1}.".format(module_name, str(e)))
            try:
                package_version = __import__(module_name[:module_name.rfind('.')] + '.version',
                                             fromlist=[module_name[:module_name.rfind('.')]])
                package_version_dict[module_name[:module_name.rfind('.')].replace('_', '-')] = package_version.__version__
            except ImportError:
                pass

            plugin_modules.extend([(name, obj) for (name, obj) in inspect.getmembers(module)
                                   if inspect.isclass(obj) and issubclass(obj, classes.SWKPlugin)])

            #second variant
            """
            package_name = entry_point.module_name
            package = __import__(package_name)
            for module in inspect.getmembers(package):
                plugin_modules.extend([(name, obj) for (name, obj) in inspect.getmembers(module)
                                           if inspect.isclass(obj) and issubclass(obj, swk_classes.SWKPlugin)])
            """


        """then sort them into command modules and parser modules"""
        plugin_command_modules.extend([(name, obj) for (name, obj) in plugin_modules
                                       if issubclass(obj, classes.SWKCommandPlugin)])
        plugin_parser_modules.extend([(name, obj) for (name, obj) in plugin_modules
                                      if issubclass(obj, classes.SWKParserPlugin)])

        logging.debug("Imported modules")
        logging.debug("All modules: {0}".format(plugin_modules))
        logging.debug("Command modules: {0}".format(plugin_command_modules))
        logging.debug("Parser modules: {0}".format(plugin_parser_modules))
        return plugin_modules, plugin_command_modules, plugin_parser_modules, package_version_dict

    def _get_available_commands_and_parsers(self):
        available_commands = dict()
        available_parsers = dict()
        for plugin_command_module in self._plugin_command_modules:
            module_name, module_class = plugin_command_module
            commands = module_class.get_commands()
            for command in commands:
                if command not in available_commands.keys():
                    available_commands[command] = module_class
                else:
                    self._die("Command {0} is defined in both plugins {1} and {2}. Ambigous definition".format(
                        command, available_commands[command].__name__, module_class.__name__))

        for plugin_parser_module in self._plugin_parser_modules:
            module_name, module_class = plugin_parser_module
            parsers = module_class.get_parsers()
            for parser in parsers:
                if parser not in available_parsers.keys():
                    available_parsers[parser] = module_class
                else:
                    self._die("Parser {0} is defined in both plugins {1} and {2}. Ambigous definition".format(
                        parser, available_parsers[parser].__name__, module_class.__name__))

        logging.debug("Available commands: {0}".format(available_commands))
        logging.debug("Available parsers: {0}".format(available_parsers))
        return available_commands, available_parsers

    def _form_commands_and_parsers_help(self):
        commands_help_message = "Commands help:\n"
        parsers_help_message = "Parsers help:\n"
        commands_help_string = ""
        parsers_help_string = ""

        for plugin_command_module in self._plugin_command_modules:
            module_name, module_class = plugin_command_module
            commands_help_message += module_class.commands_help()

        for plugin_parser_module in self._plugin_parser_modules:
            module_name, module_class = plugin_parser_module
            parsers_help_message += module_class.parsers_help()

        for k, v in self._available_commands.items():
            commands_help_string += ' ' + k

        for k, v in self._available_parsers.items():
            if k == "%":
                parsers_help_string += ' %%'  # argparse placeholders escape
            else:
                parsers_help_string += ' ' + k

        return commands_help_message, parsers_help_message, commands_help_string, parsers_help_string

    def _parse_args(self):
        argparse_epilog = '{0}'.format(self._commands_help_message) + "\n" + \
                          "{0}".format(self._parsers_help_message)

        result = dict()

        argparser = argparse.ArgumentParser(description="Swiss knife for doing everything in your infrastructure.",
                                            epilog=argparse_epilog, formatter_class=argparse.RawDescriptionHelpFormatter)
        argparser.add_argument('command', help='command. Valid choices are: {0}'.format(self._commands_help_string),
                               default='shell', nargs='?', type=str)
        argparser.add_argument('hostlist', help='hosts and/or hostgroups for command to apply to, divided by commas (,). '
                                                'Valid hostgroup modifiers are: {0}'.format(self._parsers_help_string),
                               nargs='?')
        argparser.add_argument('command_args', help='additional arguments for chosen command.', nargs=argparse.REMAINDER)
        argparser.add_argument('--version', action="version", version="%(prog)s {0}".format(self._version))

        args = argparser.parse_args(sys.argv[1:])

        result['command'] = args.command
        result['hostlist'] = args.hostlist

        result['command_args'] = args.command_args

        return result

    def _read_config(self):
        result = dict()

        config_path = self._swk_config_full_path
        if not os.path.exists(config_path):
            raise exceptions.ConfigNotFoundError("Config file not found: {0}".format(config_path))

        config = configparser.ConfigParser()
        config.read(config_path)

        for section in config.sections():
            result[section] = dict()
            for k in config.items(section):
                name, value = k
                result[section][name] = value

        return result

    def _add_empty_sections_to_config(self):
        config = self._config
        for module_name, _ in self._plugin_command_modules:
            if module_name not in config:
                config[module_name] = dict()

        for module_name, _ in self._plugin_parser_modules:
            if module_name not in config:
                config[module_name] = dict()
        logging.debug("Added missing sections to config. Now config is: {0}".format(config))
        return config

    def _find_command_executer_class(self):
        try:
            command_executer_class = self._available_commands[self._args["command"]]
        except KeyError:
            self._die("Unsupported command {0}".format(self._args["command"]))
        logging.debug("Command executer class is: {0}".format(command_executer_class))
        return command_executer_class

    def _arguments_magic(self):
        """this magic is required because hostlist is always the second parameter. if we don't need it
            we simply ignore it."""
        args = self._args
        if self._command_requires_hostlist:
            try:
                if len(args["hostlist"]) == 0:
                    self._die("Hostlist is required for command {0}, but is not provided".format(args["command"]))
            except TypeError:
                self._die("Hostlist is required for command {0}, but is not provided".format(args["command"]))
        else:
            args["command_args"].insert(0, args["hostlist"])
            args["hostlist"] = ""
        logging.debug("Did some arguments magic. Now args are: {0}".format(args))
        return args

    def _update_config(self, class_name, **kwargs):
        config = self._config[class_name]
        for k, v in kwargs.items():
            config[k] = v
        logging.debug("Updated config for class {0}. Now its config is: {1}".format(class_name, config))
        return config

    def _escape_unsafe_characters(self, string):
        return string.replace('.', r'\.')


    def _die_if_unsafe_characters(self, string):
        if string.find('*') != -1:
            self._die("Unsafe characters found in hostlist. Exiting")


    def _read_hostlist_from_stdin(self):
        hostlist_lines = sys.stdin.readlines()
        self._hostlist += '\n'.join(hostlist_lines)

    def _expand_hostlist(self):
        expanded_hostlist = list()  # expanded

        if (self._hostlist[0] == "'" and self._hostlist[-1] == "'") or \
                (self._hostlist[0] == '"' and self._hostlist[-1] == '"'):
            self._hostlist = self._hostlist[1:-1]

        # yeah, maybe we'll need that dirty hack in the future

        if len(self._hostlist) == 1 and self._hostlist[0] == '-':
            # special case for reading stdin instead of command args
            self._read_hostlist_from_stdin()

        hostgroups = self._hostlist.split()

        for hostgroup in hostgroups:
            hostlist_addition = list()
            negation = False
            hostgroup_modifier = hostgroup[0]  # the first symbol
            hostgroup_remainder = hostgroup[1:]

            if hostgroup_modifier == '-':
                negation = True
                hostgroup_modifier = hostgroup[1]
                hostgroup_remainder = hostgroup[2:]
                hostgroup = hostgroup[1:]

            if hostgroup_modifier not in self._available_parsers:
                if not hostgroup_modifier.isalnum():
                    raise exceptions.ExpandingHostlistError("Couldn't find corresponding parser for {0} modifier.".format(hostgroup_modifier))
                else:  # hostgroup is a host or a regex, not a group
                    escaped_hostgroup = self._escape_unsafe_characters(hostgroup)
                    self._die_if_unsafe_characters(escaped_hostgroup)
                    hostlist_addition = list(exrex.generate(escaped_hostgroup, limit=1000))
                    # hostlist_addition.append(hostgroup)
            else:  # we must call a parser
                parser = self._available_parsers[hostgroup_modifier]
                parser_name = parser.__name__
                self._config[parser_name] = self._update_config(parser_name, hostgroup_modifier=hostgroup_modifier,
                                                                hostgroup=hostgroup_remainder)
                obj = parser(**self._config[parser_name])
                try:
                    hostlist_addition = obj.parse()
                    if len(hostlist_addition) == 0:
                        raise exceptions.ExpandingHostlistError("Parser {0} didn't return any hosts for hostgroup {1}".format(
                            parser.__name__, hostgroup_remainder
                        ))
                except classes.SWKParsingError as e:
                    raise exceptions.ExpandingHostlistError("Parser {0} died with message: {1}".format(parser.__name__, str(e)))

            if not negation:
                # don't add host twice
                expanded_hostlist.extend([x for x in hostlist_addition if x not in expanded_hostlist])
            else:
                # delete host from list if present
                expanded_hostlist = [x for x in expanded_hostlist if x not in hostlist_addition]

        logging.debug("Expanded hostlist: {0}".format(expanded_hostlist))
        return sorted(expanded_hostlist)

    def run(self):
        if self._command == 'shell':
            if shell_mode_off:
                self._die("Please update python to python3+ to run shell mode")
            # this is a very special case
            swk.shell.SWKShellPrepare(self)
            shell = swk.shell.SWKShell(self)
            exit_status = shell.cmdloop()
            sys.exit(exit_status)

        if self._command_requires_hostlist:
            try:
                expanded_hostlist = self._expand_hostlist()
            except exceptions.ExpandingHostlistError as e:
                self._die(str(e))
        else:
            expanded_hostlist = list()

        self._config[self._command_executer_name] = self._update_config(self._command_executer_name,
                                                                        hostlist=expanded_hostlist,
                                                                        command=self._command,
                                                                        command_args=self._command_args,
                                                                        swk_dir=self._swk_dir,
                                                                        swk_path=self._swk_path,
                                                                        cwd=self._cwd,
                                                                        cache_directory=self._cache_directory_expanded)

        logging.info("Executing command with config: {0}".format(self._config[self._command_executer_name]))
        obj = self._command_executer_class(**self._config[self._command_executer_name])

        try:
            obj.run_command()
        except classes.SWKCommandError as e:
            self._die("Command class {0} died with message: {1}".format(self._command_executer_name, str(e)))
        logging.debug("swk finished")




"""
A module containing useful utility functions for swk.

swk - A tiny extendable utility for running commands against multiple hosts.

Copyright (C) 2016  Pavel "trueneu" Gurkov

see swk/main.py for more information on License and contacts
"""

import sys


class Bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'


class SWKHelperFunctions(object):
    def __init__(self):
        pass

    @staticmethod
    def print_line_with_host_prefix(line, host, is_err=False, colorful=True, print_prefix=True):
        parallel_line_prefix = '[%s]: '
        if print_prefix:
            data = parallel_line_prefix % host + line.replace('\n', '\n' + parallel_line_prefix % host, line.count('\n') - 1)
        else:
            data = line

        if not data.endswith('\n'):
            data += '\n'

        if is_err:
            if colorful:
                data = Bcolors.FAIL + data + Bcolors.ENDC
            outstream = sys.stderr
        else:
            outstream = sys.stdout

        outstream.write(data)
        outstream.flush()



"""pypi.

Desc: Library for getting information about Python packages by querying
      The CheeseShop (PYPI a.k.a. Python Package Index).


Author: Rob Cakebread <cakebread at gmail>

License  : BSD (See COPYING)

added by Pavel Gurkov (true.neu@gmail.com)
Please refer to original License text at https://github.com/cakebread/yolk
"""

from __future__ import print_function

import re
import platform
import os
import pkg_resources
if platform.python_version().startswith('2'):
    import xmlrpclib
    import urllib2
else:
    import xmlrpc.client as xmlrpclib
    import urllib.request as urllib2

XML_RPC_SERVER = 'https://pypi.python.org/pypi'


class ProxyTransport(xmlrpclib.Transport):

    """Provides an XMl-RPC transport routing via a http proxy.

    This is done by using urllib2, which in turn uses the environment
    varable http_proxy and whatever else it is built to use (e.g. the
    windows    registry).

    NOTE: the environment variable http_proxy should be set correctly.
    See check_proxy_setting() below.

    Written from scratch but inspired by xmlrpc_urllib_transport.py
    file from http://starship.python.net/crew/jjkunce/ by jjk.

    A. Ellerton 2006-07-06

    """

    def request(self, host, handler, request_body, verbose):
        """Send xml-rpc request using proxy."""
        # We get a traceback if we don't have this attribute:
        self.verbose = verbose
        url = 'http://' + host + handler
        request = urllib2.Request(url)
        try:
            request.add_data(request_body)
        except AttributeError:
            request.data = request_body
        # Note: 'Host' and 'Content-Length' are added automatically
        request.add_header('User-Agent', self.user_agent)
        request.add_header('Content-Type', 'text/xml')
        proxy_handler = urllib2.ProxyHandler()
        opener = urllib2.build_opener(proxy_handler)
        fhandle = opener.open(request)
        return self.parse_response(fhandle)


def check_proxy_setting():
    """If the environmental variable 'HTTP_PROXY' is set, it will most likely
    be in one of these forms:

    proxyhost:8080 http://proxyhost:8080 urlllib2
    requires the proxy URL to start with 'http://' This routine does that, and
    returns the transport for xmlrpc.

    """
    try:
        http_proxy = os.environ['HTTP_PROXY']
    except KeyError:
        return

    if not http_proxy.startswith('http://'):
        match = re.match(r'(http://)?([-_\.A-Za-z]+):(\d+)', http_proxy)
        os.environ['HTTP_PROXY'] = 'http://%s:%s' % (match.group(2),
                                                     match.group(3))
    return


class CheeseShop(object):

    """Interface to Python Package Index."""

    def __init__(self):
        self.xmlrpc = self.get_xmlrpc_server()

    def get_xmlrpc_server(self):
        """Return PyPI's XML-RPC server instance."""
        check_proxy_setting()
        if 'XMLRPC_DEBUG' in os.environ:
            debug = 1
        else:
            debug = 0
        return xmlrpclib.Server(XML_RPC_SERVER, transport=ProxyTransport(),
                                verbose=debug)

    def package_releases(self, package_name):
        """Query PYPI via XMLRPC interface for a pkg's available versions."""
        return self.xmlrpc.package_releases(package_name)


def get_highest_version(versions):
    """Return highest available version for a package in a list of versions.

    Uses pkg_resources to parse the versions.

    @param versions: List of PyPI package versions
    @type versions: List of strings

    @returns: string of a PyPI package version

    """
    sorted_versions = []
    for ver in versions:
        sorted_versions.append((pkg_resources.parse_version(ver), ver))

    sorted_versions = list(reversed(sorted(sorted_versions)))
    return sorted_versions[0][1]




"""
swk - A tiny extendable utility for running commands against multiple hosts.

Copyright (C) 2016  Pavel "trueneu" Gurkov

see swk/main.py for more information on License and contacts
"""

from swk import classes


class DryRunPlugin(classes.SWKCommandPlugin):
    _commands = {'dr': {'requires_hostlist': True, 'help': 'Expand the host expression and print the results. Arguments: <host exression>\n'}}
    _commands_help_message = "Dry run Plugin:\ndr - do nothing, just print hostlist\n\n"

    def __init__(self, *args, **kwargs):
        super(DryRunPlugin, self).__init__(*args, **kwargs)

    def _run(self):
        for host in self._hostlist:
            print(host)

    def run_command(self):
        self._run()



#!/usr/bin/env python
"""
swk - A tiny extendable utility for running commands against multiple hosts.

Copyright (C) 2016  Pavel "trueneu" Gurkov

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

mailto: true.neu@gmail.com
"""

import os
from swk import swiss_knife


def main():
    cwd = os.getcwd()
    swk_dir = os.path.dirname(os.path.abspath(__file__))
    swk_path = __file__
    swk_instance = swiss_knife.SwissKnife(cwd=cwd, swk_dir=swk_dir, swk_path=swk_path)
    swk_instance.run()

if __name__ == "__main__":
    main()



from setuptools import setup, find_packages

version = {}
with open('swk/version.py') as f:
    exec(f.read(), version)

url = 'https://github.com/trueneu/swiss-knife'

setup(name='swk',
      version=version['__version__'],
      description='Extendable command line utility for sysadmins',
      author="Pavel Gurkov",
      author_email="true.neu@gmail.com",
      url=url,
      download_url='https://github.com/trueneu/swiss-knife/archive/v{vers}.zip'.format(vers=version['__version__']),
      packages=find_packages(),
      license='GPLv3',
      platforms='Posix; MacOS X',
      classifiers=[
            'Development Status :: 3 - Alpha',
            'Environment :: Console',
            'Intended Audience :: Developers',
            'Intended Audience :: System Administrators',
            'Intended Audience :: End Users/Desktop',
            'Programming Language :: Python',
            'Programming Language :: Python :: 2',
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.5',
            'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
            'Topic :: System :: Systems Administration',
            'Topic :: System :: Shells',
            'Topic :: Utilities'
      ],
      keywords='cli ssh pssh swiss-knife sysadmin zabbix foreman',
      install_requires=[
          'exrex>=0.9.4',
          'paramiko>=1.16.0',
          'scp>=0.10.2',
          'pypsi>=1.3.0'
      ],
      long_description=open("README.rst").read(),
      entry_points={
          'console_scripts': [
              'swk = swk.main:main'
          ],
      },
      include_package_data=True,
      package_data={'swk': ['swk/swk.ini']},
      )



"""
swk - A tiny extendable utility for running commands against multiple hosts.

Copyright (C) 2016  Pavel "trueneu" Gurkov

see swk/main.py for more information on License and contacts
"""


from swk import classes
import paramiko
import os
import multiprocessing
import sys
import socket
import scp
import logging
import signal


class Bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'


def print_ssh_line(line, host, is_err=False, colorful=True, print_prefix=True):
    logging.debug("Received a string to print: {st}".format(st=line.encode('utf-8')))

    parallel_line_prefix = '[%s]: '
    if print_prefix:
        data = parallel_line_prefix % host + line.replace('\n', '\n' + parallel_line_prefix % host, line.count('\n') - 1)
    else:
        data = line

    if not data.endswith('\n'):
        data += '\n'

    if is_err:
        if colorful:
            data = Bcolors.FAIL + data + Bcolors.ENDC
        outstream = sys.stderr
    else:
        outstream = sys.stdout

    outstream.write(data)
    outstream.flush()


def paramiko_thread_init():
    signal.signal(signal.SIGINT, signal.SIG_IGN)


def paramiko_scp_thread_run_keyboard_interrupt_wrapper(paramiko_thread_config, source, dest):
    try:
        result = paramiko_scp_thread_run(paramiko_thread_config, source, dest)
    except KeyboardInterrupt:
        pass
    return result


def paramiko_scp_gather_thread_run_keyboard_interrupt_wrapper(paramiko_thread_config, source, dest):
    try:
        result = paramiko_scp_gather_thread_run(paramiko_thread_config, source, dest)
    except KeyboardInterrupt:
        pass
    return result


def paramiko_scp_thread_run(paramiko_thread_config, source, dest):
    connect_error_exit_code = 254
    unknown_error_exit_code = 255
    scp_error_exit_code = 253
    host = paramiko_thread_config['hostname']

    paramiko_ssh_client = paramiko.SSHClient()
    paramiko_ssh_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    try:
        paramiko_ssh_client.connect(**paramiko_thread_config)
    except socket.gaierror as e:
        print_ssh_line(e.strerror, host, is_err=True, print_prefix=True)
        return host, connect_error_exit_code
    except paramiko.SSHException as e:
        print_ssh_line(str(e), host, is_err=True, print_prefix=True)
        return host, connect_error_exit_code

    paramiko_ssh_transport = paramiko_ssh_client.get_transport()

    scpclient = scp.SCPClient(paramiko_ssh_transport)

    try:
        scpclient.put(source, dest, recursive=True)
    except OSError as e:
        print_ssh_line(str(e), host, is_err=True)
        return host, e.errno
    except scp.SCPException as e:
        print_ssh_line(str(e), host, is_err=True)
        return host, scp_error_exit_code
    except Exception as e:
        print_ssh_line(str(e), host, is_err=True)
        return host, unknown_error_exit_code

    paramiko_ssh_client.close()
    scpclient.close()
    print_ssh_line("done", host)
    return host, 0


def paramiko_scp_gather_thread_run(paramiko_thread_config, source, dest):
    connect_error_exit_code = 254
    unknown_error_exit_code = 255
    scp_error_exit_code = 253
    host = paramiko_thread_config['hostname']

    paramiko_ssh_client = paramiko.SSHClient()
    paramiko_ssh_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    try:
        paramiko_ssh_client.connect(**paramiko_thread_config)
    except socket.gaierror as e:
        print_ssh_line(e.strerror, host, is_err=True, print_prefix=True)
        return host, connect_error_exit_code
    except paramiko.SSHException as e:
        print_ssh_line(str(e), host, is_err=True, print_prefix=True)
        return host, connect_error_exit_code
    paramiko_ssh_transport = paramiko_ssh_client.get_transport()

    scpclient = scp.SCPClient(paramiko_ssh_transport)

    source_basename = os.path.basename(source)

    try:
        scpclient.get(source, local_path="{0}_{1}".format(source_basename, host), recursive=True)
    except OSError as e:
        print_ssh_line(str(e), host, is_err=True)
        return host, e.errno
    except scp.SCPException as e:
        print_ssh_line(str(e), host, is_err=True)
        return host, scp_error_exit_code
    except Exception as e:
        print_ssh_line(str(e), host, is_err=True)
        return host, unknown_error_exit_code

    paramiko_ssh_client.close()
    scpclient.close()
    print_ssh_line("done", host)
    return host, 0


def paramiko_exec_thread_run_keyboard_interrupt_wrapper(paramiko_thread_config, cmd, timeout):
    try:
        result = paramiko_exec_thread_run(paramiko_thread_config, cmd, timeout)
    except KeyboardInterrupt:
        pass
    return result


def paramiko_exec_thread_run(paramiko_thread_config, cmd, timeout):
    connect_error_exit_code = 254
    host = paramiko_thread_config['hostname']

    paramiko_ssh_client = paramiko.SSHClient()
    paramiko_ssh_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    try:
        paramiko_ssh_client.connect(**paramiko_thread_config)
    except (socket.gaierror, socket.error) as e:
        print_ssh_line(str(e), host, is_err=True, print_prefix=True)
        return host, connect_error_exit_code
    except paramiko.SSHException as e:
        print_ssh_line(str(e), host, is_err=True, print_prefix=True)
        return host, connect_error_exit_code
    paramiko_ssh_transport = paramiko_ssh_client.get_transport()
    paramiko_channel = paramiko_ssh_transport.open_session()
    paramiko_channel.exec_command(cmd)

    recv_buffer = str()
    recv_stderr_buffer = str()
    break_next_time = False

    paramiko_channel.settimeout(timeout)
    while True:
        if paramiko_channel.exit_status_ready():  # we should know if we're finished to read all the output below
            break_next_time = True

        if paramiko_channel.recv_ready():
            recv_buffer += paramiko_channel.recv(4096).decode()
        if paramiko_channel.recv_stderr_ready():
            recv_stderr_buffer += paramiko_channel.recv_stderr(4096).decode()

        if break_next_time:  # we should read until it's all in the buffer as we'll have no further opportunity
            while True:
                recv = paramiko_channel.recv(4096).decode()
                recv_buffer += recv
                if len(recv) == 0:
                    break

            while True:
                recv_stderr = paramiko_channel.recv_stderr(4096).decode()
                recv_stderr_buffer += recv_stderr
                if len(recv_stderr) == 0:
                    break

        if len(recv_buffer) > 0:
            logging.debug("Current recv_buffer: {0}".format(recv_buffer.encode()))
            last_newline_pos = recv_buffer.rfind('\n')
            if last_newline_pos != -1:
                if last_newline_pos != len(recv_buffer) - 1:
                    recv_buffer_tail = recv_buffer[last_newline_pos + 1:]
                    recv_buffer_head = recv_buffer[:last_newline_pos]
                    print_ssh_line(recv_buffer_head, host, is_err=False)
                    recv_buffer = recv_buffer_tail
                    logging.debug("recv_buffer_tail: {0}".format(recv_buffer_tail.encode()))
                    logging.debug("recv_buffer_head: {0}".format(recv_buffer_head.encode()))
                else:
                    print_ssh_line(recv_buffer, host, is_err=False)
                    recv_buffer = ""

        if len(recv_stderr_buffer) > 0:
            last_newline_pos = recv_stderr_buffer.rfind('\n')
            if last_newline_pos != -1:
                if last_newline_pos != len(recv_stderr_buffer) - 1:
                    recv_stderr_buffer_tail = recv_stderr_buffer[last_newline_pos:]
                    recv_stderr_buffer_head = recv_stderr_buffer[:last_newline_pos]
                    print_ssh_line(recv_stderr_buffer_head, host, is_err=True)
                    recv_stderr_buffer = recv_stderr_buffer_tail
                else:
                    print_ssh_line(recv_stderr_buffer, host, is_err=True)
                    recv_stderr_buffer = ""

        if break_next_time:
            break

    paramiko_ssh_client.close()
    return host, paramiko_channel.recv_exit_status()


class SSHPluginError(classes.SWKCommandError):
    def __init__(self, message):
        super(SSHPluginError, self).__init__(message)


class SSHPlugin(classes.SWKCommandPlugin):
    _commands = {'ssh': {'requires_hostlist': True, 'help': 'Executes a command over ssh host by host. Arguments: <host expression> <command to execute>\n'},
                 'pssh': {'requires_hostlist': True, 'help': 'Executes a command over ssh in parallel fashion. Arguments: <host expression> <command to execute>\n'},
                 'dist': {'requires_hostlist': True, 'help': 'Distributes a file among hosts over ssh. Arguments: <host expression> <file to distribute> <destination path>\n'
                                                             'By default, destination path is . (cwd)\n'},
                 'gather': {'requires_hostlist': True, 'help': 'Gathers remote files on local host over ssh. Arguments: <host expression> <file to gather> <destination path>\n'
                                                               'By default, destination path is . (cwd). gather adds a suffix to each file gathered to distinguish one file\n'
                                                               ' from another'}}
    _commands_help_message = "SSH plugin:\nssh - execute a command over ssh host by host (cmd)\n" \
                             "pssh - parallel exec (cmd)\n" \
                             "dist - distribute " \
                             "a file over ssh to hosts (source [destination, default is cwd])\ngather - gather " \
                             "remote files from hosts to local machine (source [destination, default is cwd]). " \
                             "gather adds a suffix (_hostname) to local filenames\n\n"

    def __init__(self, *args, **kwargs):
        super(SSHPlugin, self).__init__(*args, **kwargs)
        os.chdir(self._cwd)
        self._ssh_command = ""

        if len(self._command_args) == 0:
            raise SSHPluginError("Insufficient arguments.")

        for command_arg in self._command_args:
            self._ssh_command += command_arg + ' '

        self._ssh_command = self._ssh_command[:-1]

        if self._command == 'dist':
            if len(self._command_args) > 1:
                self._source = self._command_args[:-1]
                self._dest = self._command_args[-1]

                self._source = [os.path.expanduser(x) for x in self._source]
            else:
                self._source = self._command_args[0]
                self._dest = '.'

                self._source = os.path.expanduser(self._source)

        if self._command == 'gather':
            self._source = self._command_args[0]
            if len(self._command_args) == 2:
                self._dest = os.path.expanduser(self._command_args[1])
                if os.path.exists(self._dest):
                    if not os.path.isdir(self._dest):
                        raise SSHPluginError("gather command destination can be a directory only.")
                else:
                    try:
                        os.mkdir(self._dest)
                    except OSError as e:
                        raise SSHPluginError("couldn't mkdir {0}: {1}".format(self._dest, str(e)))
            elif len(self._command_args) == 1:
                self._dest = '.'
            else:
                raise SSHPluginError("gather command supports only two args.")

        self._hosts = self._hostlist
        self._timeout = int(getattr(self, "_timeout", 5))
        self._threads_count = int(getattr(self, "_threads_count", 10))

    def _update_paramiko_configs_from_ssh_config(self):
        #  paramiko_configs is a dict: { 'hostname1': {'hostname': hostname1, 'username': username1, etc}, ... }
        ssh_config_path = os.path.expanduser("~/.ssh/config")
        paramiko_ssh_config = paramiko.SSHConfig()

        if os.path.exists(ssh_config_path):
            with open(ssh_config_path) as f:
                paramiko_ssh_config.parse(f)

        for host_config in self._paramiko_configs:
            user_config_for_host = paramiko_ssh_config.lookup(host_config['hostname'])
            for k2 in ('hostname', 'user', 'port', 'identityfile'):
                if k2 in user_config_for_host:
                    if k2 == 'user':
                        host_config['username'] = user_config_for_host[k2]
                    elif k2 == 'identityfile':
                        host_config['key_filename'] = user_config_for_host[k2][0]
                    else:
                        host_config[k2] = user_config_for_host[k2]

    def _paramiko_configs_update_from_swk_config(self, self_attr, target_attr):
        for host_config in self._paramiko_configs:
            #  k is a host, v is a dict of settings for the host
            host_config[target_attr] = getattr(self, self_attr)


    def _paramiko_configs_set(self):
        # the way the ssh configuration for making the connections is defined is here

        self._paramiko_configs = list()

        # first, just fill in the hostnames
        for host in self._hosts:
            host_properties = dict()
            host_properties['hostname'] = host
            host_properties['timeout'] = self._timeout
            self._paramiko_configs.append(host_properties)

        # then use what user has in ssh config
        self._update_paramiko_configs_from_ssh_config()

        # then override everything with what's passed to us in constructor
        if hasattr(self, "_username"):
            self._paramiko_configs_update_from_swk_config("_username", "username")
        if hasattr(self, "_identityfile"):
            self._identityfile = os.path.expanduser(self._identityfile)
            self._paramiko_configs_update_from_swk_config("_identityfile", "key_filename")

    def _print_results_summary(self):
        failed_hosts = ""
        failed_hosts_no_comma = ""
        failed_hosts_present = False

        for result in self._results:
            host, exit_status = result
            if exit_status not in self._exit_statuses.keys():
                self._exit_statuses[exit_status] = host
            else:
                self._exit_statuses[exit_status] += ", " + host
            if exit_status != 0:
                failed_hosts += host + ", "
                failed_hosts_no_comma += host + " "
                failed_hosts_present = True

        failed_hosts_no_comma = failed_hosts_no_comma[:-1]
        if failed_hosts_present:
            sys.stderr.write("\n===\nNon-0 exit status: host; (it's not always an error)\n")
            for k, v in self._exit_statuses.items():
                if k != 0:
                    sys.stderr.write("%s: %s;\n" % (str(k), str(v)))
            called_from_shell = getattr(self, '_called_from_shell', False)
            if called_from_shell:
                fix_command = "\nRetry: {0} \"{1}\" {2}\n".format(self._command, failed_hosts_no_comma,
                                                                self._ssh_command)
            else:
                fix_command = "\nRetry: {0} {1} \"{2}\" \"{3}\"\n".format("swk", self._command,
                                                                        failed_hosts_no_comma, self._ssh_command)

            sys.stderr.write(fix_command)

    def _run(self):
        self._paramiko_configs_set()
        self._exit_statuses = dict()

        if self._command == "pssh":
            self._pool = multiprocessing.Pool(processes=min(self._threads_count, len(self._paramiko_configs)))
            self._pool_results = [self._pool.apply_async(paramiko_exec_thread_run_keyboard_interrupt_wrapper,
                                                         (paramiko_thread_config, self._ssh_command, self._timeout))
                                  for paramiko_thread_config in self._paramiko_configs]


            self._pool.close()
            try:
                self._pool.join()
                self._results = [result.get(0xFFFF) for result in self._pool_results]
            except KeyboardInterrupt:
                #print("Ctrl-C caught!")
                raise SSHPluginError("Ctrl-C caught!")
                #sys.exit(2)

        elif self._command == "ssh":  # sequential
            counter = 1
            count = len(self._hosts)
            self._results = list()
            try:
                for paramiko_config in self._paramiko_configs:
                    sys.stdout.write("%s [%d/%d]\n" % (paramiko_config["hostname"], counter, count))
                    sys.stdout.flush()
                    self._results.append(paramiko_exec_thread_run(paramiko_config, self._ssh_command, self._timeout))

                    counter += 1
            except KeyboardInterrupt:
                raise SSHPluginError("Ctrl-C caught!")
                #print("Ctrl-C caught!")
                #sys.exit(2)

        elif self._command == "dist":
            self._pool = multiprocessing.Pool(processes=self._threads_count)
            self._pool_results = [self._pool.apply_async(paramiko_scp_thread_run_keyboard_interrupt_wrapper, (paramiko_thread_config, self._source,
                                                                                   self._dest))
                                  for paramiko_thread_config in self._paramiko_configs]

            self._pool.close()
            self._pool.join()
            try:
                self._results = [result.get() for result in self._pool_results]
            except KeyboardInterrupt:
                #print("Ctrl-C caught!")
                raise SSHPluginError("Ctrl-C caught!")
                #sys.exit(2)

        elif self._command == "gather":
            os.chdir(self._dest)

            self._pool = multiprocessing.Pool(processes=self._threads_count)
            self._pool_results = [self._pool.apply_async(paramiko_scp_gather_thread_run_keyboard_interrupt_wrapper, (paramiko_thread_config,
                                                                                          self._source,
                                                                                          self._dest))
                                  for paramiko_thread_config in self._paramiko_configs]

            self._pool.close()
            self._pool.join()

            try:
                self._results = [result.get() for result in self._pool_results]
            except KeyboardInterrupt:
                #print("Ctrl-C caught!")
                raise SSHPluginError("Ctrl-C caught!")
                #sys.exit(2)

        self._print_results_summary()

    def run_command(self):
        self._run()



"""
A module containing definitions for base plugin classes for swk.

swk - A tiny extendable utility for running commands against multiple hosts.

Copyright (C) 2016  Pavel "trueneu" Gurkov

see swk/main.py for more information on License and contacts
"""
import abc
import shlex
import logging

class SWKPlugin(object):
    """
    A base class in SWK plugins hierarchy.

    This class is never used neither directly nor when defining plugins.

    """
    __metaclass__ = abc.ABCMeta

    def __init__(self, *args, **kwargs):
        """
        Constructor:
        just writes all the kwargs passed to class attributes.
        """
        for k, v in kwargs.items():
            setattr(self, "_{0}".format(k), v)


class SWKCommandPlugin(SWKPlugin):
    """
    A class used to build command plugins.

    You must derive from this class for your plugin to be recognized as a command executer.

    _commands is class attr, a dict of dicts with the following structure:
    {"command_name1" : {"requires_hostlist": True/False, "help": str},
     "command_name2" : {"requires_hostlist": True/False, "help": str}
     ...
     }
    where "requires_hostlist" is whether command works with hosts
    "help" is a help message which is shown in shell mode when user issues 'help <command_name>' command.

    _commands_help_message is a class attr, a str
    which should contain plugin's name and a brief description of all the commands included
    """
    _commands = dict()  #: this
    _commands_help_message = ""

    @staticmethod
    def _shlex_quoting_split(string):
        lex = shlex.shlex(string)
        lex.quotes = "'"
        lex.whitespace_split = True
        lex.commenters = ''

        return list(lex)

    def __init__(self, *args, **kwargs):
        """
        Constructor.

        Additional object attributes will be passed when constructing the object:
          self._command contains the string with command invoked
          self._hostlist contains the expanded hostlist if it's required by the command
          self._command_args contains the remainder of command line

        everything mentioned in config file will also be available as self._key = value
        the config file section MUST be named the same as your plugin class

        special variables that might be useful:
           self._swk_dir, self._swk_path - directory and invokation path to main executable
           self._cwd - current working directory at the time main executable was invoked
           self._cache_directory - a directory where you can store anything related to your module work
        """
        super(SWKCommandPlugin, self).__init__(*args, **kwargs)
        shlex_splitted_command_args = list()
        if hasattr(self, "_command_args") and len(self._command_args) > 0 and (self._command_args[0] is not None):
            logging.debug("self._command_args received: {cmdargs}".format(cmdargs=self._command_args))
            for command_arg in self._command_args:
                shlex_splitted_command_args.extend(self._shlex_quoting_split(command_arg))
            self._command_args = shlex_splitted_command_args

    @classmethod
    def get_commands(cls):
        return cls._commands.keys()

    @classmethod
    def get_command_help(cls, command):
        return cls._commands.get(command).get('help', 'no help provided\n')

    @classmethod
    def requires_hostlist(cls, command):
        return cls._commands.get(command).get('requires_hostlist', True)

    @classmethod
    def commands_help(cls):
        return cls._commands_help_message

    @abc.abstractmethod
    def run_command(self):
        """
        This method is called when SWK has done all the preparations and needs the plugin to
        actually execute the command.

        You should redefine it.
        Here you should determine which command is called and process it.
        """


class SWKParserPlugin(SWKPlugin):
    """
    A class used to build parser plugins.

    You must derive from this class for your plugin to be recognized as a hostlist parser.

    _parsers is class attr, a dict of dicts with the following structure:
    {"hostlist_modifier1" : {"help": str},
     "hostlist_modifier2" : {"help": str},}
     ...
     }
    where "help" is a help message which is shown in shell mode when user issues 'help <hostlist_modifier>' command.

    _parsers_help_message is a class attr, a str
    which should contain a brief description of all the hostlist modifiers available
    """

    _parsers = dict()
    _parsers_help_message = ""

    def __init__(self, *args, **kwargs):
        """
        Constructor.

        Additional object attributes will be passed when constructing the object:

        self._hostgroup contains string with hostgroup to be parsed
        self._hostgroup_modifier contains the parser specifier, a single symbol

        everything mentioned in config file will also be available as self._key = value
        the config file section MUST be named the same as your plugin class
        """
        super(SWKParserPlugin, self).__init__(*args, **kwargs)

    @classmethod
    def get_parsers(cls):
        return cls._parsers.keys()

    @classmethod
    def get_parser_help(cls, parser):
        return cls._parsers.get(parser).get('help', 'no help provided\n')

    @abc.abstractmethod
    def parse(self):
        """
        This method is called when SWK needs the plugin to parse a hostlist expression.

        You should redefine it.
        Here you should determine which modifier is used and act accordingly.

        :return: Must return list() of hostnames.
        """

    @classmethod
    def parsers_help(cls):
        return cls._parsers_help_message


class SWKParsingError(Exception):
    """
    A class to derive your parsing errors from.
    Raise it if there's an unrecoverable error when parsing a host expression to stop the command from executing.
    """
    def __init__(self, message):
        """
        Constructor

        :param message: a message to be displayed
        """
        super(SWKParsingError, self).__init__(message)


class SWKCommandError(Exception):
    """
    A class to derive your command execution errors from.
    Raise it if there's an unrecoverable error when executing a command.
    """
    def __init__(self, message):
        """
        Constructor

        :param message: a message to be displayed
        """
        super(SWKCommandError, self).__init__(message)
